/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <suplib/str.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <linklist/linklist.h>

#include "aperture.h"
#include "aperture_queue.h"
#include "assembly_check.h"

#include "xmatrix.h"
#include "xstack.h"
#include "transform.h"

void
aperture_empty( Aperture *aperture)
{
  aperture->module->cleanup(aperture);
  XMatrix_free( aperture->forward );
  XMatrix_free( aperture->backward );
  aperture->forward = aperture->backward = NULL;

  free( aperture->id.label );
  aperture->id.label = NULL;
}

void
aperture_free( Aperture *aperture)
{
  aperture_empty( aperture );
  free(aperture);
}

int
aperture_init( void* aperture, void* udata )
{
  ((Aperture*)aperture)->module->init( (Aperture*) aperture, udata);
  return 0;
}

/*

  PROCEDURE
  aperture_new - create a new aperture node.

  DESCRIPTION
  This function creates a new aperture node and attaches it to the
  current assembly's aperture queue.  It records the passed data
  address in the aperture, and copies the current forward and reverse
  transformation matrices into the aperture.

  AUTHOR
  Dan Nguyen

*/
void
  aperture_new(
    ApertureModule *module,	/* the module structure */
    void *data_object, 		/* User information to
				   construct the assembly */
    const char* label           /* The optional label */
    )
{
  Aperture *aperture;                                                 

  tf_enter("aperture_new");
  
  /* ensure that we're in an assembly.  doesn't return if we're not */
  in_assembly();

  if ( NULL == ( aperture = malloc( sizeof(Aperture) ) ) )
    tf_exit(ExitERR_alloc, "unable to allocate new aperture node");
  
  aperture->forward = xstack_top_copy();
  aperture->backward = Xmatrix_inverse( aperture->forward );
  aperture->module = module;
  aperture->data_object = data_object;
  aperture->id.label = NULL;

  /* if label is specified use it.  otherwise, aperture_queue_insert
     takes care of creating a unique label */
  if ( label )
  {
    aperture->id.label = str_dup( label );
    tf_assert( NULL != aperture->id.label );
  }

  {
    size_t i;
    for ( i = 0 ; i < MAXPROCESSRESULT ; i++ )
    {
      aperture->status[i] = 0;
      aperture->weight[i] = 0.0;
    }
  }

  aperture_queue_insert( aperture );

  tf_leave();
}


int
aperture_print( Aperture *aperture )
{
  int j, k, l;

  tf_message( "Assembly # %ld, aperture # %ld, module: %s\n",
	      aperture->id.assembly, aperture->id.aperture,
	      aperture->module->name );

  tf_message( "forward:\t\t\tbackward:\n" );
  for ( j = 0; j < DIM4; j++ ) { /* loop thru rows */
    tf_message( "" );
    for ( l = 0; l < 2; l++ ) {	/* forward, backward */
      for ( k = 0; k < DIM4; k++ ) /* loop tru columns */
	tf_message( " %-6.3f", 
		l == 0 ? aperture->forward[j][k] :aperture->backward[j][k] );
      tf_message( "\t" );
    }
    tf_message( "\n" );
  }

  return 0;
}


void
aperture_cp( Aperture* from, Aperture* to )
{
  *to = *from;
}
