/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <suplib/str.h>
#include <linklist/linklist.h>

#include "init.h"

#include "assembly_queue.h"
#include "aperture.h"
#include "aperture_queue.h"

/*
   the current aperture id.  aperture id's are unique within an
   assembly, not between.
*/
static unsigned long aperture_id = 0;

/* the leaves for the current assembly */
static SLinkList aperture_queue = NULL;

/*--------------------------------------------------------------------*/
SLinkList
aperture_queue_new( void )
{
  tf_enter( "aperture_queue_new" );

  aperture_queue = sll_new( NULL );
  if ( NULL == aperture_queue )
    tf_die( "error creating new aperture queue\n" );

  aperture_id = 0;

  tf_leave();

  return aperture_queue;
}

/*--------------------------------------------------------------------*/
void
aperture_queue_free( SLinkList queue  )
{
  tf_enter( "aperture_queue_free" );

  sll_delete( queue, (SLL_DELETE) aperture_free );

  tf_leave();
}

/*--------------------------------------------------------------------*/
void
aperture_queue_insert( Aperture *aperture )
{
  tf_enter( "aperture_queue_insert" );

  tf_assert(aperture_queue != NULL);

  aperture->id.assembly = get_assembly_id();
  aperture->id.aperture = aperture_id++;

  if ( NULL == aperture->id.label )
  {
    char buf[512];
    sprintf( buf, "a%03ld_%03ld", aperture->id.assembly, aperture->id.aperture );
    aperture->id.label = str_dup( buf );
    tf_assert( NULL != aperture->id.label );
  }


  if ( sll_insert_tail( aperture_queue, aperture ) )
    tf_exit(ExitERR_alloc, "unable to insert new aperture in aperture stack");

  tf_leave();
  
}

/*--------------------------------------------------------------------*/
int
aperture_queue_print( SLinkList queue )
{
  sll_traverse( queue, (SLL_ACTION) aperture_print );

  return sll_count( queue );
}


/*--------------------------------------------------------------------*/
/*
  PROCEDURE
  Copy an aperture node in a linklist to an array.
 */

static int
aperture_cp_node_to_array( void *node, void *udata )
{
  Aperture *from = node;
  Aperture **ap_pp = udata;
  Aperture *to = *ap_pp;

  aperture_cp( from, to );
  (*ap_pp)++;

  return 0;
}

/* create an array of apertures from a linklist of apertures */
Aperture*
aperture_queue_array_new( SLinkList queue )
{
  Aperture* array;
  Aperture* tarray;

  tf_enter( "aperture_queue_ll_to_array" );

  array = tarray = malloc( sll_count(queue) * sizeof( Aperture ) );

  if ( NULL == array )
    tf_exit( ExitERR_alloc, "Unable to allocate aperture array" );

  sll_utraverse( queue, aperture_cp_node_to_array, &tarray );

  tf_leave();

  return array;
}


/* DO NOT free the apertures in the aperture array! these are shallow
   copies of the apertures still in the aperture_queues.  those are
   dealt with when the actual queues are destroyed
*/
void
aperture_queue_array_free( Aperture queue[] )
{
  free( queue );
}
