/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <linklist/linklist.h>

#include "assembly.h"
#include "assembly_check.h"

static SLinkList check_assembly=NULL;

static void
print_del_check_assembly( void *data )
{
  AssemblyType *ass = data;

  tf_message( "\tBegin %s\n" , assembly_get_name(*ass) );

  free( ass );
}

/*
  PROCEDURE
    Print diagnostic messages, perform a stack dump then exit.

  DESCRIPTION
    Warn user what is expected to build the assembly.
    Print out the stack of assembly that has been requested so far.

  AUTHOR
    Dan Nguyen    
*/
static void
  clean_up( AssemblyType type )
{
  SLLNode node;
  AssemblyType *ass;

  tf_enter("clean_up");

  node = sll_head_node( check_assembly );
  tf_assert(node != NULL);

  tf_message( "Expected end_" );
  ass = sll_node_get_data( node );
  tf_assert(ass != NULL);
  
  assembly_print_type( *ass );
  fputs( ", got end_" , stderr );

  assembly_print_type( type );
  fputs( " instead!\n", stderr);

  tf_message("The assembly stack:\n");

  sll_delete(check_assembly, print_del_check_assembly );

  tf_die( "Out of sequence" );
}

void
  assembly_enter( AssemblyType type )
{
  AssemblyType *ass;

  tf_enter("assembly_enter");
  
  if ( type == ASSEMBLY )
  {
    if ( check_assembly != NULL )
      tf_die( "assemblies cannot be nested - missing call to end_assembly" );

    if ( (check_assembly = sll_new( NULL )) == NULL )
      tf_exit( ExitERR_alloc, "Unable to allocate check_assembly" );
  }
  else if ( check_assembly == NULL )
    tf_die( "must call begin_assembly() first" );
    
  if ( NULL == ( ass = malloc( sizeof(AssemblyType) ) ) )
    tf_exit(ExitERR_alloc, "unable to allocate assembly check list node");

  *ass = type;
  if ( sll_insert_head( check_assembly, ass ) )
    tf_exit(ExitERR_alloc, "unable to insert node into assembly check list");
    
  tf_leave();
  return;
}

void
  assembly_leave( AssemblyType type )
{
  AssemblyType *ass;
  SLLNode node;

  tf_enter("assembly_leave");

  if ( check_assembly == NULL )
    tf_die( "lua function end_%s called outside of an assembly",
	    assembly_get_name(type));

  node = sll_head_node( check_assembly );
  tf_assert(node != NULL);
  ass =  sll_node_get_data( node );
  tf_assert(ass != NULL);

  /*   if user check out the assembly in the wrong order, then exit */
  if ( *ass != type )
    clean_up( type );		/* doesn't return from this */

  free( ass );
  sll_destroy_node( check_assembly, node );

  /* if checking out an ASSEMBLY, the list of assmblies should be empty */
  if (type == ASSEMBLY)
  {
    if ( sll_count( check_assembly ) )
      tf_die( "lua function end_assembly called out of sequence");
    sll_delete(check_assembly, NULL);
    check_assembly = NULL;
  }

  tf_leave();
  return;
}

void
  check_assemblies( void )
{
  tf_enter("check_assemblies");

  if ( check_assembly != NULL )
  {
    tf_message( "premature ending of the assembly!\n");
    tf_message("The assembly stack:\n");
    
    sll_delete(check_assembly, print_del_check_assembly );

    tf_die( "one or more missing end_(sub)assembly calls" );
  }

  tf_leave();
}

/*
  this routine returns if an assembly exists, else it exits.  useful
  for the aperture modules.
*/
void
in_assembly( void )
{
  if ( check_assembly == NULL )
    tf_die( "must call begin_assembly() before calling a module" );
}

/*
  this routine returns true if an assembly exists, else false.
*/
int
is_in_assembly( void )
{
  return check_assembly != NULL;
}
