/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include "config.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>


#include "luav.h"

#include <suplib/str.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "params.h"

#include "lua_register.h"

#include "aperture.h"
#include "assembly_check.h"
#include "xstack.h"

#include "common.h"

#include "lua_iface.h"

/*---------------------------------------------------------------------*/

static char *
stack_top_msg( lua_State *L )
{
  const char *topmsg;
  char *msg;

  tf_assert( lua_gettop( L ) && (topmsg = lua_tostring(L, 1)) );
  tf_assert( msg = str_dup( topmsg ) );
  return msg;
}

lua_State * register_lua( Params *io )
{
    lua_State *L;
    tf_enter("register_lua");

    /* create interpreter instance */
    if ( NULL == ( L = luaL_newstate() ) )
	tf_die( "error creating Lua interpreter" );

    /* register standard lua libraries */
    luaL_openlibs( L );
  
    /* register aperture definition functions */
    {
	ApertureModule **mod;
	for (mod = modules ; *mod ; mod++ )
	    lua_register( L, (*mod)->name, (*mod)->function);
    }

    /* register other systems */
    assembly_queue_register_lua( L );
    transform_register_lua( L );
    xstack_register_lua( L );

    return L;
}

/*---------------------------------------------------------------------*/

void exec_lua(  Params *io, lua_State *L )
{
  /* set up override function, if specified */
    if ( '\0' == io->override[0] ) 
    {
	if ( luaL_dostring( L, "function override()\n end" ) )
	    tf_die( "internal error creating stub override: %s", stack_top_msg( L )  );
    }

    else if ( '@' == io->override[0] )
    {
	/* an '@' as the first letter signifies a file name follows */
	if ( luaL_dofile( L, &io->override[1] ) )
	{
	    tf_die( "Error while parsing override function in file `%s': %s",
		    &io->override[1], stack_top_msg( L ) );
	}
    }

  else
  {
    char filename[ 9 ];
    int fd;
    FILE *file = NULL;

    strcpy(filename, "apXXXXXX" );

    fd = mkstemp( filename );
    if ( -1 == fd )
      tf_die( "unable to create source file for override() function\n" );
    

    file = fdopen( fd, "w" );
    if ( NULL == file )
    {
      close(fd);
      unlink( filename );
      tf_die( "error getting file handle for override() file\n" );
    }
	 
    fprintf( file, "function override()\n%s\nend",
	     io->override);
    fclose( file );

    if ( luaL_dofile( L, filename ) )
    {
	unlink( filename );
	tf_die( "Error while creating override function: %s", stack_top_msg( L ) );
    }

    unlink( filename );
  }

    if ( luaL_dofile( L, io->lua ) )
      tf_die( "Error while executing definition file %s: %s", io->lua, stack_top_msg( L ) );

    lua_close( L );

  /* make sure that things were setup correctly */
  check_assemblies( );

  xstack_check( );

  tf_leave();
}
