/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

#include "luav.h"

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>

#include "module.h"
#include "ray_transforms.h"
#include "aperture.h"
#include "common.h"

#include "luasup.h"

/* forward declaration of function local to this file */
static int strut( lua_State *L );
static void init(Aperture *aperture, void* ami);
static ProcessResult process(Aperture *aperture);
static void cleanup(Aperture *aperture);

/*
   Define the ApertureModule structure for this module.
*/
ApertureModule strut_module =
{
  "strut",		/* the lua name of the instantiation function */
  strut,		/* the C lua-callable instantiation function */
  init,			/* the aperture initialization function */
  process,		/* the aperture photon processing function */
  cleanup		/* the aperture cleanup function */
};


/*
   this structure contains all of the information necessary
   to describe an instantiation of this aperture.
*/

typedef struct {

  double half_width;
  int    interior;
  ProcessResult result;

} Strut;


/*
   this structure contains the offsets into the data packet of
   the data required by this aperture.  it is initialized by
   the init() function in this file.
*/
   
static struct
{
  DVector3 *position;
  DVector3 *direction;
} photon;

/*
   this function is called from lua to create an instantiation of
   this aperture type.
*/
static int
strut( lua_State *L ) {

    Strut  *data;			/* do not free */
    double   width;
    int narg = 0;
    const char  *label = NULL;

    tf_enter("strut");

    if ( NULL == (data = malloc( sizeof(Strut) )) )
	tf_exit( ExitERR_alloc, "unable to allocate data for strut" );
    
    width = luaL_checknumber( L, ++narg );
    luaL_argcheck( L, width > 0.0, narg, "strut width must be positive");

    data->half_width = width / 2.0;
    data->interior     = luaL_checkint( L, ++narg );
    data->result       = luaL_checkint( L, ++narg ) ? PASSED : BLOCKED;

    label = get_opt_label( L, narg );
    if ( label ) narg++ ;

    luaL_argcheck( L, lua_gettop(L) == narg, ++narg, "too many arguments" );

    aperture_new( &strut_module, data, label );

    tf_leave();

    return 0;
}

/*

   This routine is called once, before any photons arrive.  It should
   be used to determine offsets to whatever data this aperture needs
   from the photon data packet.  It should also initialize anything
   that needs initialization.
   
*/

static void 
init(Aperture *aperture, void* amip)
{
  ApertureModuleInfo* ami = amip;
  DpktField *dpktf_pos, *dpktf_dir;

  static int initialized = 0;

  /* do this only once to save some CPU */
  if (! initialized)
  {
    tf_enter("strut init");

    if ( NULL == (dpktf_pos = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_pos )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_pos
	                           "' in input stream" );

    if ( NULL == (dpktf_dir = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_dir )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_dir
	                           "' in input stream" );

    photon.position  = (DVector3 *) bpipe_dpktf_data( dpktf_pos, ami->dpkt );
    photon.direction = (DVector3 *) bpipe_dpktf_data( dpktf_dir, ami->dpkt );

    initialized = 1;

    tf_leave();
  }
}

/*

   This routine is called for every photon.  It should be as efficient
   as possible.
   
*/

static ProcessResult
process( Aperture *aperture )
{

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */
  Strut *data = aperture->data_object;

  /* apply transformation to the photon */
  TRANSFORM_POSITION ( aperture->forward, photon.position, &p );
  TRANSFORM_DIRECTION( aperture->forward, photon.direction, &d );

  /* translate it to the z = 0 plane */
  PROJECT_RAY(&p, &d, 0.0);

  if (data->interior)
  {
    return ( p.x >= -data->half_width &&
	     p.x <=  data->half_width
	   )
      ? data->result : NOTAPPLICABLE;
  }
  else
  {
    return ( p.x < -data->half_width ||
	     p.x >  data->half_width
	   )
      ? data->result : NOTAPPLICABLE;
  }
}

/*
   This routine is called once for every instantiatian of this aperture.
   It should free instantiation specific resources (like those allocated
   in the lua-called instantiation creation routine).
*/

static void
cleanup (Aperture *aperture)
{
  free(aperture->data_object);
}
